/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : neumannBW.cpp                                 |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal                               |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Acoustical scattering using H-Matrix BEM      |
 |  `---'  |                solver for Neumann boundary condition         |
 +========================================================================+=+
 */

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>
#include <castor/hmatrix.hpp>
#include <castor/linalg.hpp>
#include <castor/graphics.hpp>
#include "fem.hpp"
#include "bem.hpp"

using namespace castor;

int main (int argc, char* argv[])
{
    // Load meshes
    matrix<double> Svtx, Pvtx;
    matrix<size_t> Stri, Ptri;
    std::tie(Stri,Svtx) = triread("/Users/aussal/Git/leprojetcastor/fembem/demo/data/","unitSphere_1e3.ply");
    std::tie(Ptri,Pvtx) = triread("/Users/aussal/Git/leprojetcastor/fembem/demo/data/","Oxyz_1e4.ply");
    
    // Graphical representation
    figure fig;
    trimesh(fig,Ptri,Pvtx);
    trimesh(fig,Stri,Svtx);
    
    // Parameters
    matrix<double> U = {-1,0,0};
    double k = 5.85;
    float tol = 1e-3;
    
    // FEM
    tic();
    femdata<double> v(Stri,Svtx,lagrangeP1,3);
    femdata<double> u(Stri,Svtx,lagrangeP1,3);
    toc();

    // Incident wave : PW(x,u) = exp(i*k*x.u)
    auto Pinc = planeWave(Pvtx,U,k);
    
    // Identity matrix
    tic();
    auto Id = mass<std::complex<double>>(v);
    toc();
    
    // Coupling coeff
    std::complex<double> beta(0,k);

    // Left hand side : [1i*k*beta*(-Id/2 + Dt) - H]
    tic();
    auto LHSfct = [&v,&u,&k,&beta,&Id](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return beta*(-0.5*eval(Id(Ix,Iy)) + transpose(doubleLayer<std::complex<double>>(v,u,k,Iy,Ix)))
        - hypersingular<std::complex<double>>(v,u,k,Ix,Iy);
    };
    hmatrix<std::complex<double>> LHS(v.dof(),u.dof(),tol,LHSfct);
    toc();
    disp(LHS);

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    auto B = - rightHandSide<std::complex<double>>(v,dnPWsource,U,k);

    // Solve linear system : [1i*k*beta*(-Id/2 + Dt) - H] = -dnP0
    auto mu     = gmres(LHS,B,tol,100);
    auto lambda = beta * mu;
        
    // Boundary radiation
    tic();
    auto Sbndfct = [&v,&u,&k,&Id](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayer<std::complex<double>>(v,u,k,Ix,Iy);
    };
    hmatrix<std::complex<double>> Sbnd(v.dof(),u.dof(),tol,Sbndfct);
    auto Dbndfct = [&v,&u,&k,&Id](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return 0.5*eval(Id(Ix,Iy)) + doubleLayer<std::complex<double>>(v,u,k,Ix,Iy);
    };
    hmatrix<std::complex<double>> Dbnd(v.dof(),u.dof(),tol,Dbndfct);
    matrix<std::complex<double>> Pbnd = mtimes(Sbnd,lambda) - mtimes(Dbnd,mu);
    toc();
    Pbnd = gmres(Id,Pbnd,tol,100) + planeWave(v.dof(),U,k);
    
    // Domain radiation
    tic();
    auto Sdomfct = [&Pvtx,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return singleLayer<std::complex<double>>(Pvtx,u,k,Ix,Iy);
    };
    hmatrix<std::complex<double>> Sdom(Pvtx,u.dof(),tol,Sdomfct);
    auto Ddomfct = [&Pvtx,&u,&k](matrix<std::size_t> Ix, matrix<std::size_t> Iy)
    {
        return doubleLayer<std::complex<double>>(Pvtx,u,k,Ix,Iy);
    };
    hmatrix<std::complex<double>> Ddom(Pvtx,u.dof(),tol,Ddomfct);
    matrix<std::complex<double>> Pdom = mtimes(Sdom,lambda) - mtimes(Ddom,mu) + Pinc;
    toc();
    for (std::size_t l=0; l<size(Pvtx,1); ++l)
    {
        if (std::pow(Pvtx(l,0),2)+std::pow(Pvtx(l,1),2)+std::pow(Pvtx(l,2),2)<1.1) {Pdom(l)=Pinc(l);}
    }
    
    // Graphical representation
    figure fig2;
    trimesh(fig2,Stri,Svtx,abs(Pbnd));
    trimesh(fig2,Ptri,Pvtx,abs(Pdom));
    
    // Infinite radiation
    tic();
    matrix<double> theta = transpose(linspace(0,2*M_PI,1e3));
    matrix<double> Nu = horzcat(horzcat(cos(theta),sin(theta)),zeros(size(theta)));
    matrix<std::complex<double>> Pinf(numel(theta),1);
    for (std::size_t i=0; i<numel(theta); ++i)
    {
        auto Sinf = singleLayerInfinite<std::complex<double>>(Nu,u,k,i);
        auto Dinf = doubleLayerInfinite<std::complex<double>>(Nu,u,k,i);
        Pinf(i) = (std::complex<double>) (mtimes(Sinf,lambda) - mtimes(Dinf,mu));
    }
    toc();
    
    // Graphical representation
    figure fig3;
    plot(fig3,theta,transpose(log(abs(Pinf))),{"-b"});
    
    // Plot
    drawnow(fig);
    disp("done !");
    return 0;
}
