/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : overview_bem.cpp                              |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal & Marc Bakry                  |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Boundary Elements Methods for standard        |
 |  `---'  |                operators S, D, Dt, H in dense format         |
 +========================================================================+
 */

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>
#include <castor/linalg.hpp>
#include <castor/graphics.hpp>
#include "fem.hpp"
#include "bem.hpp"

using namespace castor;

int main(int argc, char* argv[])
{
    //===============================================================
    std::cout << "+===================+" << std::endl;
    std::cout << "|       MESHES      |" << std::endl;
    std::cout << "+===================+" << std::endl;

    // Sphere mesh
    matrix<double> X, Y, Z, Svtx, Pvtx;
    matrix<size_t> Stri, Ptri;
    std::tie(X,Y,Z) = sphere2(1000);
    X = horzcat(0,reshape(X,1,numel(X)));
    Y = horzcat(0,reshape(Y,1,numel(Y)));
    Z = horzcat(0,reshape(Z,1,numel(Z)));
    std::tie(Stri,Svtx) = tetdelaunay(X,Y,Z);
    std::tie(Stri,Svtx) = tetboundary(Stri,Svtx);
    
    // Planar mesh
    std::tie(X,Y) = meshgrid(linspace(-4,4,50));
    std::tie(Ptri,Pvtx) = tridelaunay(X,Y,zeros(size(X)));

    // Graphical representation
    figure fig;
    trimesh(fig,Stri,Svtx);
    trimesh(fig,Ptri,Pvtx);
        
    //===============================================================
    std::cout << "+=======================+" << std::endl;
    std::cout << "|      SINGLE LAYER     |" << std::endl;
    std::cout << "+=======================+" << std::endl;
    
    // FEM
    femdata<double> v(Stri,Svtx,lagrangeP1,3);
    femdata<double> u(Stri,Svtx,lagrangeP1,3);
    
    // Incident wave : PW(x,u) = exp(i*k*x.u)
    matrix<double> U = {-1,0,0};
    double k = 5;
    auto Pinc = planeWave(Pvtx,U,k);

    // Identity matrix
    tic();
    auto Id = full(mass<std::complex<double>>(v));
    toc();

    // Left hand side : \int_sigma_x \int_sigma_y v(x) G(x,y) u(y) dx dy
    tic();
    auto LHS = singleLayer<std::complex<double>>(v,u,k);
    toc();
    disp(sum(LHS)); //   -0.6614 + 2.3185i

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    tic();
    auto RHS = - rightHandSide<std::complex<double>>(v,PWsource,U,k);
    toc();

    // Solve linear system : S lambda = -PO
    tic();
    auto lambda = linsolve(LHS,RHS);
    toc();

    // Radiation
    tic();
    auto Pbnd = linsolve(Id,mtimes(LHS,lambda)) + planeWave(v.dof(),U,k);
    toc();
    tic();
    auto Sdom = singleLayer<std::complex<double>>(Pvtx,u,k);
    auto Pdom = mtimes(Sdom,lambda) + Pinc;
    toc();

    // Graphical representation
    figure fig2;
    trimesh(fig2,Ptri,Pvtx,abs(Pdom));
    trimesh(fig2,Stri,Svtx,abs(Pbnd));

    //===============================================================
    std::cout << "+=======================+" << std::endl;
    std::cout << "|      DOUBLE LAYER     |" << std::endl;
    std::cout << "+=======================+" << std::endl;

    // Left hand side : \int_sigma_x \int_sigma_y v(x) dnyG(x,y) u(y) dx dy
    tic();
    LHS = -(0.5*Id + doubleLayer<std::complex<double>>(v,u,k));
    toc();
    disp(sum(LHS)); //    -1.6068 + 5.6271i

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    RHS = - rightHandSide<std::complex<double>>(v,PWsource,U,k);

    // Solve -(Id/2 - D) = -P0
    auto mu = linsolve(LHS,RHS);

    // Boundary radiation
    Pbnd = linsolve(Id,mtimes(LHS,mu)) + planeWave(v.dof(),U,k);

    // Domain radiation
    auto Ddom = doubleLayer<std::complex<double>>(Pvtx,u,k);
    Pdom = - mtimes(Ddom,mu) + Pinc;
    for (std::size_t l=0; l<size(Pvtx,1); ++l)
    {
        if (std::pow(Pvtx(l,0),2)+std::pow(Pvtx(l,1),2)+std::pow(Pvtx(l,2),2)<1.1) {Pdom(l)=Pinc(l);}
    }

    // Graphical representation
    figure fig3;
    trimesh(fig3,Ptri,Pvtx,abs(Pdom));
    trimesh(fig3,Stri,Svtx,abs(Pbnd));

    //===============================================================
    std::cout << "+========================+" << std::endl;
    std::cout << "| DOUBLE LAYER TRANSPOSE |" << std::endl;
    std::cout << "+========================+" << std::endl;

    // Left hand side : \int_sigma_x \int_sigma_y v(x) dnyG(x,y) u(y) dx dy
    tic();
    LHS = -0.5*Id + transpose(doubleLayer<std::complex<double>>(v,u,k));
    toc();
    disp(sum(LHS)); //

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    RHS = - rightHandSide<std::complex<double>>(v,dnPWsource,U,k);

    // Solve (-Id/2 + Dt) = -Pdn0
    lambda = linsolve(LHS,RHS);

    // Boundary radiation
    auto Sbnd = singleLayer<std::complex<double>>(v,u,k);
    Pbnd = linsolve(Id,mtimes(Sbnd,lambda)) + planeWave(v.dof(),U,k);

    // Domain radiation
    Sdom = singleLayer<std::complex<double>>(Pvtx,u,k);
    Pdom = mtimes(Sdom,lambda) + Pinc;
    for (std::size_t l=0; l<size(Pvtx,1); ++l)
    {
        if (std::pow(Pvtx(l,0),2)+std::pow(Pvtx(l,1),2)+std::pow(Pvtx(l,2),2)<1.1) {Pdom(l)=Pinc(l);}
    }

    // Graphical representation
    figure fig4;
    trimesh(fig4,Stri,Svtx,abs(Pbnd));
    trimesh(fig4,Ptri,Pvtx,abs(Pdom));

    //===============================================================
    std::cout << "+===================+" << std::endl;
    std::cout << "|   HYPERSINGULAR   |" << std::endl;
    std::cout << "+===================+" << std::endl;

    // Left hand side : k^2 * \int_Sx \int_Sy n.psi(x) G(x,y) n.psi(y) dx dy - \int_Sx \int_Sy nxgrad(psi(x)) G(x,y) nxgrad(psi(y)) dx dy
    tic();
    LHS = - hypersingular<std::complex<double>>(v,u,k);
    toc();
    disp(sum(LHS)); //     -22.4022 -13.6573i

    // Right hand side : - \int_sigma_x v(x) PW(x,u) dx
    RHS = - rightHandSide<std::complex<double>>(v,dnPWsource,U,k);

    // Solve -H = -dnP0
    mu = linsolve(LHS,RHS);

    // Boundary radiation
    auto Dbnd = 0.5*Id + doubleLayer<std::complex<double>>(v,u,k);
    Pbnd = -linsolve(Id,mtimes(Dbnd,mu)) + planeWave(v.dof(),U,k);

    // Domain radiation
    Ddom = doubleLayer<std::complex<double>>(Pvtx,u,k);
    Pdom = - mtimes(Ddom,mu) + Pinc;
    for (std::size_t l=0; l<size(Pvtx,1); ++l)
    {
        if (std::pow(Pvtx(l,0),2)+std::pow(Pvtx(l,1),2)+std::pow(Pvtx(l,2),2)<1.1) {Pdom(l)=Pinc(l);}
    }

    // Graphical representation
    figure fig5;
    trimesh(fig5,Stri,Svtx,abs(Pbnd));
    trimesh(fig5,Ptri,Pvtx,abs(Pdom));
    
    //===============================================================
    std::cout << "+====================+" << std::endl;
    std::cout << "|     SUB-MATRIX     |" << std::endl;
    std::cout << "+====================+" << std::endl;
    
    // Collocation
    matrix<std::size_t> I = colon(30,2,900);
    matrix<std::size_t> J = colon(900,-3,7);
    matrix<double> S  = singleLayer<double>(Svtx,u,0.);
    matrix<double> Sb = singleLayer<double>(Svtx,u,0.,I,J);
    disp(norm(eval(S(I,J))-Sb,"inf"));
    
    // Galerkin
    S  = singleLayer<double>(v,u,0.);
    Sb = singleLayer<double>(v,u,0.,I,J);
    disp(norm(eval(S(I,J))-Sb,"inf"));
    
    //===============================================================
    std::cout << "+===================+" << std::endl;
    std::cout << "|      DRAWNOW      |" << std::endl;
    std::cout << "+===================+" << std::endl;
    
    drawnow(fig);    
    
    disp("done !");
    return 0;
}
