/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : fem.hpp                                       |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal & Marc Bakry                  |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Quadrature and Finite Element basic tools     |
 |  `---'  |                                                              |
 +========================================================================+
 */

#pragma once

#define CASTOR_FEM_HPP

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>

namespace castor
{

//==========================================================================//
//                           ENVIRONMENT DATA                               //
//==========================================================================//
enum femtype{vertices, lagrangeP0, lagrangeP1};


//==========================================================================//
//                        TRIANGLEDATA CLASS                                //
//==========================================================================//
//==========================================================================
// [triangledata]
///
template<typename T>
class triangledata
{
public:
    auto intsing(T x0, T y0, T z0) const;
    bool isfar(T x0, T y0, T z0) const;
    void projectP1(T x0, T y0, T z0, matrix<T>& P) const;
    void update(matrix<T>const& S);
    void update(matrix<T>const& A, matrix<T>const& B, matrix<T>const& C);
    
    matrix<T>const& ctr() const {return m_ctr;};
    matrix<T>const& nrm() const {return m_nrm;};
    matrix<T>const& nu()  const {return m_nu;};
    matrix<T>const& tau() const {return m_tau;};
    T const& A(std::size_t i) const   {return m_A(i);};
    T const& B(std::size_t i) const   {return m_B(i);};
    T const& C(std::size_t i) const   {return m_C(i);};
    T const& ctr(std::size_t i) const {return m_ctr(i);};
    T const& hgt(std::size_t i) const {return m_hgt(i);};
    T const& nrm(std::size_t i) const {return m_nrm(i);};
    T const& nu(std::size_t i, std::size_t j) const {return m_nu(i,j);};
    T const& srf() const              {return m_srf;}
    T const& tau(std::size_t i, std::size_t j) const {return m_tau(i,j);};
    
private:
    matrix<T> m_A   = matrix<T>(1,3); // FIRST NODE
    matrix<T> m_B   = matrix<T>(1,3); // SECOND NODE
    matrix<T> m_C   = matrix<T>(1,3); // THIRD NODE
    matrix<T> m_ctr = matrix<T>(1,3); // CENTER
    matrix<T> m_hgt = matrix<T>(1,3); // HEIGHT
    matrix<T> m_lgt = matrix<T>(1,3); // EDGE LENGTH
    matrix<T> m_nrm = matrix<T>(1,3); // TRIANGLE NORMAL
    matrix<T> m_nu  = matrix<T>(3,3); // EDGE NORMALS
    T         m_srf = 0;              // SURFACE
    T         m_stp = 0;              // MAX EDGE LENGTH
    matrix<T> m_tau = matrix<T>(3,3); // EDGE TANGENT
};

//==========================================================================
//
///
template<typename T>
auto triangledata<T>::intsing(T x0, T y0, T z0) const
{
    // initialize
    std::size_t ip, ipp;
    matrix<T> xX0S(1,3), yX0S(1,3), zX0S(1,3);
    matrix<T> nrmX0S(1,3), ca(1,3), sa(1,3), xnu(1,3);
    T omega, h, ps, psp1, ps2, ah, ar, intaRm1;
    T Rm1, dnRm1;
    T tol = 1e-7;
    
    // Point to vertex
    xX0S(0) = m_A(0) - x0;
    xX0S(1) = m_B(0) - x0;
    xX0S(2) = m_C(0) - x0;
    
    yX0S(0) = m_A(1) - y0;
    yX0S(1) = m_B(1) - y0;
    yX0S(2) = m_C(1) - y0;
    
    zX0S(0) = m_A(2) - z0;
    zX0S(1) = m_B(2) - z0;
    zX0S(2) = m_C(2) - z0;

    for(std::size_t i=0; i<3; ++i)
    {
        nrmX0S(i) = std::sqrt(xX0S(i)*xX0S(i)+yX0S(i)*yX0S(i)+zX0S(i)*zX0S(i));
    }
    
    // Point height to triangle
    h = xX0S(0)*m_nrm(0) + yX0S(0)*m_nrm(1) + zX0S(0)*m_nrm(2);
    
    // solid angle
    if (std::abs(h)<tol) {omega = 0;}
    else
    {
        for(std::size_t i=0; i<3; ++i)
        {
            ip = (i+1)%3;
            ca(i) = (xX0S(i)*xX0S(ip)+yX0S(i)*yX0S(ip)+zX0S(i)*zX0S(ip))/(nrmX0S(i)*nrmX0S(ip));
            sa(i) = std::sqrt(1 - ca(i)*ca(i));
        }
        omega = -M_PI;
        for(std::size_t i=0; i<3; ++i)
        {
            ip = (i+1)%3; ipp = (ip+1)%3;
            omega += std::acos((ca(i)-ca(ip)*ca(ipp))/(sa(ip)*sa(ipp)));
        }
        if (h < 0) {omega = -omega;}
    }
    if (std::isnan(omega))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Point is close to triangle, change h tolerance.");
    }
    
    // initialization of integration
    Rm1 = -h*omega;
    dnRm1 = -omega;
    
    // edge integration
    for(std::size_t i=0; i<3; ++i)
    {
        ip  = (i+1)%3;
        ipp = (ip+1)%3;
        // projections
        ps   = xX0S(ip)*m_tau(i,0)  + yX0S(ip)*m_tau(i,1)  + zX0S(ip)*m_tau(i,2);
        psp1 = xX0S(ipp)*m_tau(i,0) + yX0S(ipp)*m_tau(i,1) + zX0S(ipp)*m_tau(i,2);
        ps2  = ps*(nrmX0S(ipp) - nrmX0S(ip));
        // Subtil
        ar = std::abs(psp1 - ps);
        xnu(0) = xX0S(ip) - ps*m_tau(i,0);
        xnu(1) = yX0S(ip) - ps*m_tau(i,1);
        xnu(2) = zX0S(ip) - ps*m_tau(i,2);
        ah = std::sqrt(xnu(0)*xnu(0) + xnu(1)*xnu(1) + xnu(2)*xnu(2));
        // integration of 1/R on the edge - singular case
        if( nrmX0S(ip)-std::abs(ps) < tol)
        {
            intaRm1 = 0;
            // Particle on the left of the edge
            if (ps > tol && psp1 > tol) {intaRm1 = std::log(nrmX0S(ipp)/nrmX0S(ip));}
            // Particle on the right of the edge
            else if (ps < -tol && psp1 < -tol) {intaRm1 = -std::log(nrmX0S(ipp)/nrmX0S(ip));}
        }
        // integration of 1/R on the edge - general case h>tol
        else
        {
            intaRm1 = std::asinh(psp1/ah) - std::asinh(ps/ah);
        }
        // 1/r
        Rm1 += intaRm1*(xX0S(ip)*m_nu(i,0) + yX0S(ip)*m_nu(i,1) + zX0S(ip)*m_nu(i,2));
        // grad[1/r]
        dnRm1 += intaRm1*(m_nu(i,0)*m_nrm(0) + m_nu(i,1)*m_nrm(1) + m_nu(i,2)*m_nrm(2));
    }
    return std::make_tuple(Rm1, dnRm1);
}

//==========================================================================
//
///
template<typename T>
bool triangledata<T>::isfar(T x0, T y0, T z0) const
{
    return ((x0-m_ctr(0))*(x0-m_ctr(0)) +
            (y0-m_ctr(1))*(y0-m_ctr(1)) +
            (z0-m_ctr(2))*(z0-m_ctr(2)) ) > 1.1*m_stp;
}

//==========================================================================
//
///
template<typename T>
void triangledata<T>::projectP1(T x0, T y0, T z0, matrix<T>& P) const
{
    // ((Sk-X) dot NUj) / Hj
    P(0) = ((m_B(0)-x0)*m_nu(0,0) +
            (m_B(1)-y0)*m_nu(0,1) +
            (m_B(2)-z0)*m_nu(0,2) ) / m_hgt(0);
    P(1) = ((m_C(0)-x0)*m_nu(1,0) +
            (m_C(1)-y0)*m_nu(1,1) +
            (m_C(2)-z0)*m_nu(1,2) ) / m_hgt(0);
    P(2) = ((m_A(0)-x0)*m_nu(2,0) +
            (m_A(1)-y0)*m_nu(2,1) +
            (m_A(2)-z0)*m_nu(2,2) ) / m_hgt(2);
}

//==========================================================================
//
///
template<typename T>
inline void triangledata<T>::update(matrix<T>const& S)
{
    update({S(0,0),S(0,1),S(0,2)},{S(1,0),S(1,1),S(1,2)},{S(2,0),S(2,1),S(2,2)});
}

//==========================================================================
//
///
template<typename T>
inline void triangledata<T>::update(matrix<T>const& A, matrix<T>const& B, matrix<T>const& C)
{
    // Nodes and center
    for(std::size_t j=0; j<3; ++j)
    {
        m_A(j)   = A(j);
        m_B(j)   = B(j);
        m_C(j)   = C(j);
        m_ctr(j) = (m_A(j)+m_B(j)+m_C(j))/3.;
    }
    
    // Edges
    matrix<T> AB = B-A;
    matrix<T> BC = C-B;
    matrix<T> CA = A-C;
    
    // Normal to ABC and surface
    m_nrm(0) = AB(1)*BC(2) - AB(2)*BC(1);
    m_nrm(1) = AB(2)*BC(0) - AB(0)*BC(2);
    m_nrm(2) = AB(0)*BC(1) - AB(1)*BC(0);
    m_srf    = 0.5*std::sqrt(m_nrm(0)*m_nrm(0) + m_nrm(1)*m_nrm(1) + m_nrm(2)*m_nrm(2));
    m_nrm   /= 2*m_srf;
    
    // length edges
    m_lgt(2) = std::sqrt(AB(0)*AB(0)+AB(1)*AB(1)+AB(2)*AB(2));
    m_lgt(0) = std::sqrt(BC(0)*BC(0)+BC(1)*BC(1)+BC(2)*BC(2));
    m_lgt(1) = std::sqrt(CA(0)*CA(0)+CA(1)*CA(1)+CA(2)*CA(2));
    m_stp    = std::max(std::max(m_lgt(0),m_lgt(1)),m_lgt(2));
    
    // tangent to edges
    for(std::size_t j=0; j<3; ++j)
    {
        m_tau(2,j) = AB(j)/m_lgt(2);
        m_tau(0,j) = BC(j)/m_lgt(0);
        m_tau(1,j) = CA(j)/m_lgt(1);
    }
    
    // normal to edges
    for(std::size_t i=0; i<3; ++i)
    {
        m_nu(i,0) = m_tau(i,1)*m_nrm(2)-m_tau(i,2)*m_nrm(1);
        m_nu(i,1) = m_tau(i,2)*m_nrm(0)-m_tau(i,0)*m_nrm(2);
        m_nu(i,2) = m_tau(i,0)*m_nrm(1)-m_tau(i,1)*m_nrm(0);
    }
    
    // Height
    m_hgt(0) = AB(0)*m_nu(0,0) + AB(1)*m_nu(0,1) + AB(2)*m_nu(0,2);
    m_hgt(1) = BC(0)*m_nu(1,0) + BC(1)*m_nu(1,1) + BC(2)*m_nu(1,2);
    m_hgt(2) = CA(0)*m_nu(2,0) + CA(1)*m_nu(2,1) + CA(2)*m_nu(2,2);
};


//==========================================================================//
//                           FEMDATA CLASS                                  //
//==========================================================================//
//==========================================================================
// [femdata]
///
template<typename T>
class femdata
{
public:
    femdata(matrix<std::size_t>const& elt, matrix<T>const& vtx, femtype fet, std::size_t ngs=0);
    auto dofsubdata(matrix<std::size_t>const& Idof) const;
    auto refQuadrature() const;
    
    matrix<T>const& dof() const    {return m_dof;};
    matrix<T>const& gss() const    {return m_gss;};
    matrix<T>const& gssnrm() const {return m_gssnrm;};
    matrix<T>const& wgt() const    {return m_wgt;};
    
    matrix<std::size_t>const& elt2dof() const {return m_elt2dof;};
    matrix<std::size_t>const& elt2gss() const {return m_elt2gss;};
    
    T const& base(std::size_t i, std::size_t j) const {return m_base(i,j);};
    T const& dof(std::size_t i, std::size_t j) const {return m_dof(i,j);};
    T const& grad(std::size_t i, std::size_t j, std::size_t k) const {return m_grad[k](i,j);};
    T const& gss(std::size_t i, std::size_t j) const {return m_gss(i,j);};
    T const& gssnrm(std::size_t i, std::size_t j) const {return m_gssnrm(i,j);};
    T const& nxgrad(std::size_t i, std::size_t j, std::size_t k) const {return m_nxgrad[k](i,j);};
    T const& wgt(std::size_t i) const {return m_wgt(i);};
    
    std::size_t const& elt2dof(std::size_t i, std::size_t j) const {return m_elt2dof(i,j);};
    std::size_t const& elt2gss(std::size_t i, std::size_t j) const {return m_elt2gss(i,j);};
    
    triangledata<T>const& tridata(std::size_t i) const {return m_tridata[i];};
    
private:
    // PRIVATE BUILDER
    void m_buildBase();
    void m_buildDof();
    void m_buildGeometry();
    void m_buildGrad();
    void m_buildQuadrature();
    
    // ATTRIBUTS
    femtype                          m_fet;       // ENUM       : TYPE OF FINITE ELEMENT
    std::size_t                      m_ngs;       // DIMENSION  : SIZE OF ELEMENTARY QUADRATURE
    matrix<T>                        m_dof;       // COORDINATES: DEGREES OF FREEDOM
    matrix<T>                        m_gss;       // COORDINATES: GAUSSIAN QUADRATURE
    matrix<T>                        m_gssnrm;    // COORDINATES: NORMALS AT GAUSS POINTS
    matrix<T>                        m_vtx;       // COORDINATES: NODES
    matrix<T>                        m_wgt;       // VALUES     : GAUSSIAN WEIGHTS
    std::vector<matrix<std::size_t>> m_dof2elt;   // TABLE      : DOF TO ELEMENT
    matrix<std::size_t>              m_elt2dof;   // TABLE      : ELEMENTS TO DOF
    matrix<std::size_t>              m_elt2gss;   // TABLE      : ELEMENTS TO GAUSSIAN QUADRATURE
    matrix<std::size_t>              m_elt2vtx;   // TABLE      : ELEMENTS TO NODES
    matrix<T>                        m_base;      // OPERATOR   : BASIS
    std::vector<matrix<T>>           m_grad;      // OPERATOR   : GRADIENT
    std::vector<matrix<T>>           m_nxgrad;    // OPERATOR   : NORMALxGRADIENT
    std::vector<triangledata<T>>     m_tridata;   // GEOMETRY   : TRIANGLE DATA
};

//==========================================================================
//
///
template<typename T>
femdata<T>::femdata(matrix<std::size_t>const& elt, matrix<T>const& vtx, femtype fet, std::size_t ngs)
{
    // Check input
    if (size(vtx,2)!=3 || size(elt,2)!=3)
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Vertices should be a Nvtx-by-3 real matrix and elements a Nelt-by-3 integer matrix.");
    }
    
    // Input data
    m_vtx     = vtx;
    m_elt2vtx = elt;
    m_ngs     = ngs;
    m_fet     = fet;
    
    // Computed data
    m_buildGeometry();
    m_buildQuadrature();
    m_buildDof();
    m_buildBase();
    m_buildGrad();
}

//==========================================================================
//
///
template<typename T>
auto femdata<T>::dofsubdata(matrix<std::size_t>const& Idof) const
{
    // Elements associated to dof indices
    std::vector<std::size_t> v;
    for (std::size_t l=0; l<numel(Idof); ++l)
    {
        v.insert(v.end(),std::begin(m_dof2elt[Idof(l)].val()),std::end(m_dof2elt[Idof(l)].val()));
    }
    matrix<std::size_t> Ielt = unique<std::size_t>(v);
    
    // Local renumerotation of dof indices
    matrix<std::size_t> e2d = eval(m_elt2dof(Ielt,col(m_elt2dof)));
    matrix<std::size_t> i1=argsort(Idof), i2=argsort(e2d);
    matrix<long> e2dl(size(e2d,1),size(e2d,2),-1);
    std::size_t k=0, l=0;
    while (l<numel(e2d))
    {
        if (e2d(i2(l))<Idof(i1(k))) {++l;}
        else if (e2d(i2(l))==Idof(i1(k)))
        {
            e2dl(i2(l)) = i1(k);
            ++l;
        }
        else if (e2d(i2(l))>Idof(i1(k)))
        {
            ++k;
            if (k==numel(Idof)){break;}
        }
    }
    return std::make_tuple(Ielt,e2dl);
}

//==========================================================================
//
///
template<typename T>
auto femdata<T>::refQuadrature() const
{
    matrix<T> gssloc(m_ngs,2), wgtloc(m_ngs,1);
    double a, b, w;
    switch(m_ngs)
    {
        case 0:
            break;
            
        case 1:
            gssloc(0,0) = 1/3.; gssloc(0,1) = 1/3.; wgtloc(0) = 0.5;
            break;
            
        case 3:
            a = 2/3.; b = 1/6.; w = 1/6.;
            wgtloc(0) = w; wgtloc(1) = w; wgtloc(2) = w;
            gssloc(0,0) = a; gssloc(0,1) = b;
            gssloc(1,0) = b; gssloc(1,1) = a;
            gssloc(2,0) = b; gssloc(2,1) = b;
            break;
            
        case 6:
            a = 0.816847572980459; b = 0.091576213509771; w = 0.109951743655322/2.;
            wgtloc(0) = w; wgtloc(1) = w; wgtloc(2) = w;
            gssloc(0,0) = a; gssloc(0,1) = b;
            gssloc(1,0) = b; gssloc(1,1) = a;
            gssloc(2,0) = b; gssloc(2,1) = b;
            a = 0.108103018168070; b = 0.445948490915965; w = 0.223381589678011/2.;
            wgtloc(3) = w; wgtloc(4) = w; wgtloc(5) = w;
            gssloc(3,0) = a; gssloc(3,1) = b;
            gssloc(4,0) = b; gssloc(4,1) = a;
            gssloc(5,0) = b; gssloc(5,1) = b;
            break;
            
        default:
            error(__FILE__, __LINE__, __FUNCTION__,"Unvalid quadrature size (available are 0, 1, 3 and 6).");
            break;
    }
    return std::make_tuple(gssloc,wgtloc);
}

//==========================================================================
//
///
template<typename T>
void femdata<T>::m_buildBase()
{
    // Initialize
    matrix<T> gssloc,wgtloc;
    std::vector<matrix<T>> fct(size(m_elt2dof,2),matrix<T>(1,m_ngs));
    m_base = zeros<T>(size(m_gss,1),size(m_elt2dof,2));
    
    // Reference quadrature
    std::tie(gssloc,wgtloc) = refQuadrature();
    
    // Basis functions
    switch(m_fet)
    {
        case vertices :
            break;
            
        case lagrangeP0:
            fct[0] = ones<T>(1,m_ngs);
            break;
            
        case lagrangeP1:
            for(std::size_t g=0; g<m_ngs; ++g)
            {
                fct[0](g) = 1 - gssloc(g,0) - gssloc(g,1);
                fct[1](g) = gssloc(g,0);
                fct[2](g) = gssloc(g,1);
            }
            break;
            
        default:
            error(__FILE__, __LINE__, __FUNCTION__,"Invalid basis function type");
            break;
    }
    
    // Assembling
    for (std::size_t e=0; e<size(m_elt2vtx,1); ++e)
    {
        for (std::size_t d=0; d<size(m_elt2dof,2); ++d)
        {
            for (std::size_t g=0; g<m_ngs; ++g)
            {
                m_base(m_elt2gss(e,g),d) = fct[d](g);
            }
        }
    }
}

//==========================================================================
//
///
template<typename T>
void femdata<T>::m_buildDof()
{
    // Dof, elt2dof
    switch(m_fet)
    {
        case vertices :
            break;
            
        case lagrangeP0 :
            m_dof = zeros<T>(size(m_elt2vtx,1),size(m_vtx,2));
            for (std::size_t i=0; i<size(m_elt2vtx,1); ++i)
            {
                for (std::size_t j=0; j<size(m_elt2vtx,2); ++j)
                {
                    m_dof(i,0) += m_vtx(m_elt2vtx(i,j),0);
                    m_dof(i,1) += m_vtx(m_elt2vtx(i,j),1);
                    m_dof(i,2) += m_vtx(m_elt2vtx(i,j),2);
                }
            }
            m_dof /= size(m_elt2vtx,2);
            m_elt2dof = transpose(range(0,size(m_elt2vtx,1)));
            break;
            
        case lagrangeP1 :
            m_dof     = m_vtx;
            m_elt2dof = m_elt2vtx;
            break;
            
        default:
            error(__FILE__, __LINE__, __FUNCTION__,"Invalid basis function type");
            break;
    }
    
    // Number of elements close to dof
    matrix<std::size_t> ind(1,size(m_dof,1));
    for (std::size_t l=0; l<numel(m_elt2dof); ++l)
    {
        ++ind(m_elt2dof(l));
    }
    
    // Initialize tables
    m_dof2elt.resize(size(m_dof,1));
    for (std::size_t d=0; d<size(m_dof,1); ++d)
    {
        m_dof2elt[d].resize(1,ind(d));
        ind(d)=0;
    }
    
    // Fill tables
    std::size_t d;
    for (std::size_t e=0; e<size(m_elt2dof,1); ++e)
    {
        for (std::size_t j=0; j<size(m_elt2dof,2); ++j)
        {
            d = m_elt2dof(e,j);
            m_dof2elt[d](ind(d)) = e;
            ++ind(d);
        }
    }
    
    // Check error
    if (sum(ind)!=numel(m_elt2dof))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Unavailable case.");
    }
}

//==========================================================================
//
///
template<typename T>
void femdata<T>::m_buildGrad()
{
    // Initialize
    matrix<T> gssloc,wgtloc;
    std::vector<matrix<T>> fctX(size(m_elt2dof,2),zeros<T>(1,m_ngs));
    std::vector<matrix<T>> fctY(size(m_elt2dof,2),zeros<T>(1,m_ngs));
    m_grad   = std::vector<matrix<T>>(3, matrix<T>(size(m_gss,1),size(m_elt2dof,2)));
    m_nxgrad = std::vector<matrix<T>>(3, matrix<T>(size(m_gss,1),size(m_elt2dof,2)));
    matrix<T> AB(1,3), AC(1,3), G(2,2), Gm1(2,2), P(2,3);
    T detG;
    
    // Reference quadrature
    std::tie(gssloc,wgtloc) = refQuadrature();
    
    // Basis functions
    switch (m_fet)
    {
        case vertices :
            break;
            
        case lagrangeP0:
            break;
            
        case lagrangeP1:
            for(std::size_t g=0; g<m_ngs; ++g)
            {
                fctX[0](g) = -1; fctY[0](g) = -1;
                fctX[1](g) = 1;  fctY[1](g) = 0;
                fctX[2](g) = 0;  fctY[2](g) = 1;
            }
            break;
            
        default:
            error(__FILE__, __LINE__, __FUNCTION__,"Invalid basis function type");
            break;
    }
    
    // Build grad operator
    for(std::size_t e=0; e<size(m_elt2vtx,1); ++e)
    {
        // For each coordinate
        G = zeros<T>(2,2);
        for(std::size_t j=0; j<3; ++j)
        {
            // Vector basis
            AB(j) = m_vtx(m_elt2vtx(e,1),j) - m_vtx(m_elt2vtx(e,0),j);
            AC(j) = m_vtx(m_elt2vtx(e,2),j) - m_vtx(m_elt2vtx(e,0),j);
            
            // Gramm matrix (symetric)
            G(0,0) += AB(j)*AB(j);
            G(0,1) += AB(j)*AC(j);
            G(1,1) += AC(j)*AC(j);
        }
        G(1,0) = G(0,1);
        
        // inverse of Gramm matrix
        detG     = G(0,0)*G(1,1) - G(1,0)*G(0,1);
        Gm1(0,0) = G(1,1)/detG;  Gm1(0,1) = -G(0,1)/detG;
        Gm1(1,0) = -G(1,0)/detG; Gm1(1,1) = G(0,0)/detG;
        
        // P = (G^{-1})^t {AB;AC}
        for(std::size_t j=0; j<3; ++j)
        {
            P(0,j) = Gm1(0,0)*AB(j) + Gm1(1,0)*AC(j);
            P(1,j) = Gm1(0,1)*AB(j) + Gm1(1,1)*AC(j);
        }
        
        // grad(PHI) = P * grad(PHI_ref)
        for(std::size_t j=0; j<3; ++j)
        {
            for(std::size_t d=0; d<size(m_elt2dof,2); ++d)
            {
                for(std::size_t g=0; g<m_ngs; ++g)
                {
                    m_grad[j](m_elt2gss(e,g),d) = P(0,j)*fctX[d](g) + P(1,j)*fctY[d](g);
                }
            }
        }
    }
    
    // Build nxgrad operator
    for(std::size_t k=0; k<m_grad.size(); ++k)
    {
        for(std::size_t i=0; i<size(m_grad[k],1); ++i)
        {
            for(std::size_t j=0; j<size(m_grad[k],2); ++j)
            {
                m_nxgrad[k](i,j) = (m_gssnrm(i,(k+1)%3)*m_grad[(k+2)%3](i,j) -
                                    m_gssnrm(i,(k+2)%3)*m_grad[(k+1)%3](i,j) );
            }
        }
    }
}

//==========================================================================
//
///
template<typename T>
void femdata<T>::m_buildGeometry()
{
    m_tridata = std::vector<triangledata<T>>(size(m_elt2vtx,1),triangledata<T>());
    matrix<T> A(1,3), B(1,3), C(1,3);
    for (std::size_t e=0; e<size(m_elt2vtx,1); ++e)
    {
        for (std::size_t j=0; j<3; ++j)
        {
            A(j) = m_vtx(m_elt2vtx(e,0),j);
            B(j) = m_vtx(m_elt2vtx(e,1),j);
            C(j) = m_vtx(m_elt2vtx(e,2),j);
        }
        m_tridata[e].update(A,B,C);
    }
}

//==========================================================================
//
///
template<typename T>
void femdata<T>::m_buildQuadrature()
{
    // Initialize data
    std::size_t ne = size(m_elt2vtx,1);
    m_gss          = zeros<T>(m_ngs*ne,3);
    m_wgt          = zeros<T>(m_ngs*ne,1);
    m_gssnrm       = zeros<T>(m_ngs*ne,3);
    m_elt2gss      = zeros<std::size_t>(ne,m_ngs);
    matrix<T> AB(1,3), AC(1,3), gssloc, wgtloc;
    std::size_t ind;
    
    // Quadrature on reference element {(0,0), (1,0), (0,1)}
    std::tie(gssloc,wgtloc) = refQuadrature();
    
    // For each element
    for(std::size_t e=0; e<ne; ++e)
    {
        // Basis vectors (AB and AC)
        for(std::size_t j=0; j<3; ++j)
        {
            AB(j) = m_tridata[e].B(j) - m_tridata[e].A(j);
            AC(j) = m_tridata[e].C(j) - m_tridata[e].A(j);
        }
        
        // Local gauss nodes: XG = A + [B-A C-A]*XGloc, weights and normals
        for(std::size_t g=0; g<m_ngs; ++g)
        {
            ind = e*m_ngs + g;
            m_elt2gss(e,g) = ind;
            m_wgt(ind)     = 2*m_tridata[e].srf()*wgtloc(g);
            for(std::size_t j=0; j<3; ++j)
            {
                m_gss(ind,j)    = m_tridata[e].A(j) + AB(j)*gssloc(g,0) + AC(j)*gssloc(g,1);
                m_gssnrm(ind,j) = m_tridata[e].nrm(j);
            }
        }
    }
}


//==========================================================================//
//                        VARIATIONAL FORMULATIONS                          //
//==========================================================================//
//==========================================================================
//
///
template<typename T, typename S>
smatrix<T> mass(femdata<S>const& v)
{
    std::vector<std::size_t> ind;
    std::vector<T> val;
    std::size_t ndof = size(v.dof(),1);
    for (std::size_t e=0; e<size(v.elt2dof(),1); ++e)
    {
        for (std::size_t g=0; g<size(v.elt2gss(),2); ++g)
        {
            std::size_t ig = v.elt2gss(e,g);
            for (std::size_t i=0; i<size(v.elt2dof(),2); ++i)
            {
                for (std::size_t j=0; j<size(v.elt2dof(),2); ++j)
                {
                    ind.push_back(v.elt2dof(e,i)*ndof+v.elt2dof(e,j));
                    val.push_back(v.base(ig,i) * v.wgt(ig) * v.base(ig,j));
                }
            }
        }
    }
    return smatrix<T>(ndof,ndof,ind,val);
}

//==========================================================================
//
///
template<typename T, typename S>
smatrix<T> rigidity(femdata<S>const& v)
{
    std::vector<std::size_t> ind;
    std::vector<T> val;
    std::size_t ndof = size(v.dof(),1);
    T tmp;
    for (std::size_t e=0; e<size(v.elt2dof(),1); ++e)
    {
        for (std::size_t g=0; g<size(v.elt2gss(),2); ++g)
        {
            std::size_t ig = v.elt2gss(e,g);
            for (std::size_t i=0; i<size(v.elt2dof(),2); ++i)
            {
                for (std::size_t j=0; j<size(v.elt2dof(),2); ++j)
                {
                    ind.push_back(v.elt2dof(e,i)*ndof+v.elt2dof(e,j));
                    tmp = v.wgt(ig) *
                    (  v.grad(ig,i,0) * v.grad(ig,j,0)
                     + v.grad(ig,i,1) * v.grad(ig,j,1)
                     + v.grad(ig,i,2) * v.grad(ig,j,2) );
                    val.push_back(tmp);
                }
            }
        }
    }
    return smatrix<T>(ndof,ndof,ind,val);
}

}

